# Targeted Maximum Likelihood Estimation
# for non-parametric estimation of the marginal effect of a binary point
# treatment, adjusting for treatment (g) and missingness (g_Delta) mechanisms
# Parameters include: additive treatment effect: E_W[E(Y|A=1,W) - E(Y|A=0,W)]
# and, for binary outcomes, relative risk(RR) and odds ratio(RR)
#   mu1 = E_W[E(Y|A=1,W)], mu0 = E_W[E(Y|A=0,W)]
#   RR = mu1/mu0
#   OR =  mu1/(1-mu1)/(mu0 / (1-mu0))
# author: Susan Gruber, sgruber@berkeley.edu
# date:   September 25, 2010
# 
# Models or estimated values for Q, g, g_Delta can be user-supplied or 
# estimated using super learner.  Cross-validated inital Q can be obtained
# by specifying cvQinit=TRUE

# References: 
#  M.J. van der Laan, S. Rose, S. Gruber. Readings in Targeted Maximum Likelihood Estimation.
#  U.C. Berkeley Division of Biostatistics Working Paper Series, Working Paper 254.
#  http://www.bepress.com/ucbbiostat/paper254

#-------------summary.tmle------------------
# create tmle summary object
# return values of psi, variance, epsilon, 
# and models for Q, g, g_Delta, if available
#---------------------------------------	 
summary.tmle <- function(object,...) {
  if(identical(class(object), "tmle")){
		Qmodel <- Qcoef <- gmodel <- gcoef <- g_Deltamodel <- g_Deltacoef <- NULL
		Qterms <- gterms <- g_Deltaterms <- ""
		if(!is.null(object$Qinit)){
		  if (!is.null(object$Qinit$coef)) {
			Qcoef <- object$Qinit$coef	
			if(class(Qcoef) == "matrix"){
				Qterms <- colnames(Qcoef)
			} else {
				Qterms <- names(Qcoef)
			}
			Qmodel <- paste("Y ~ 1")
			if(length(Qterms) > 1) {
				Qmodel <- paste("Y ~ 1 +", paste(Qterms[-1], collapse =" + "))
			} 
		  } else {
		  	Qmodel <- object$Qinit$type
		  }
		}
		if(!is.null(object$g)){
		if (!is.null(object$g$coef)) {
			gcoef <- object$g$coef
			if(class(gcoef) == "matrix"){
				gterms <- colnames(gcoef)
			} else {
				gterms <- names(gcoef)
			}			
			gmodel <- paste("Y ~ 1")
			if(length(gterms) > 1) {
				gmodel <- paste("Y ~ 1 +", paste(gterms[-1], collapse =" + "))
			} 
		}}
		if(!is.null(object$g_Delta)){
		if (!is.null(object$g_Delta$coef)) {
			g_Deltacoef <- object$g_Delta$coef
			if(class(g_Deltacoef) == "matrix"){
				g_Deltaterms <- colnames(g_Deltacoef)
			} else {
				g_Deltaterms <- names(g_Deltacoef)
			}			
			g_Deltamodel <- paste("Y ~ 1")
			if(length(g_Deltaterms) > 1) {
				g_Deltamodel <- paste("Y ~ 1 +", paste(g_Deltaterms[-1], collapse =" + "))
			} 
		}}
		summary.tmle <- list(estimates=object$estimates,
						Qmodel=Qmodel, Qterms=Qterms, Qcoef=Qcoef, Qtype=object$Qinit$type, gmodel=gmodel, gterms=gterms, gcoef=gcoef,gtype=object$g$type, 
						g_Deltamodel=g_Deltamodel, g_Deltaterms=g_Deltaterms, g_Deltacoef=g_Deltacoef, g_Deltatype=object$g_Delta$type )
		class(summary.tmle) <- "summary.tmle"
	} else {
		stop("object must have class 'tmle'")
		summary.tmle <- NULL
	}
	return(summary.tmle)
}

#-------------print.summary.tmle------------------
# print tmle summary object
#-------------------------------------------------
print.summary.tmle <- function(x,...) {
  if(identical(class(x), "summary.tmle")){
  	cat(" Initial estimation of Q\n")
  	cat("\t Procedure:", x$Qtype)
  	if(!(is.na(x$Qcoef[1]))){
  		cat("\n\t Model:\n\t\t", x$Qmodel)
  		cat("\n\n\t Coefficients: \n")
  		terms <- sprintf("%15s", x$Qterms)
  		extra <- ifelse(x$Qcoef >= 0, "  ", " ")
  		for(i in 1:length(x$Qcoef)){
  			cat("\t", terms[i], extra[i], x$Qcoef[i], "\n")
  		}
  	}
  	cat("\n Estimation of g (treatment mechanism)\n")
  	cat("\t Procedure:", x$gtype,"\n")
  	if(!(is.na(x$gcoef[1]))){	
 		cat("\t Model:\n\t\t", x$gmodel, "\n")
  		cat("\n\t Coefficients: \n")
  		terms <- sprintf("%15s", x$gterms)
  		extra <- ifelse(x$gcoef >= 0, "  ", " ")
  		for(i in 1:length(x$gcoef)){
  			cat("\t", terms[i], extra[i], x$gcoef[i], "\n")
  	}} 
  	cat("\n Estimation of g_Delta (missingness mechanism)\n")
  	cat("\t Procedure:", x$g_Deltatype, "\n")
  	if(!(is.na(x$g_Deltacoef[1]))){		
  		cat("\t Model:\n\t\t",x$g_Deltamodel, "\n")
  		cat("\n\t Coefficients: \n")
  		terms <- sprintf("%15s", x$g_Deltaterms)
  		extra <- ifelse(x$g_Deltacoef >= 0, "  ", " ")
  		for(i in 1:length(x$g_Deltacoef)){
  			cat("\t", terms[i], extra[i], x$g_Deltacoef[i], "\n")
  	}}
	if(!is.null(x$estimates$ATE)){
			cat("\n Additive Effect")
			cat("\n   Parameter Estimate: ", signif(x$estimates$ATE$psi,5))
			cat("\n   Estimated Variance: ", signif(x$estimates$ATE$var.psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$ATE$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$ATE$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$ATE$CI[1],5), ", ", signif(x$estimates$ATE$CI[2],5), ")", sep=""),"\n") 
		}
		if(!is.null(x$estimates$RR)){
			cat("\n Relative Risk")
			cat("\n   Parameter Estimate: ", signif(x$estimates$RR$psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$RR$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$RR$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$RR$CI[1],5), ", ", signif(x$estimates$RR$CI[2],5), ")", sep=""),"\n") 
			cat("\n              log(RR): ", signif(x$estimates$RR$log.psi,5))
			cat("\n    variance(log(RR)): ", signif(x$estimates$RR$var.log.psi,5), "\n")

		}
		if(!is.null(x$estimates$OR)){
			cat("\n Odds Ratio")
			cat("\n   Parameter Estimate: ", signif(x$estimates$OR$psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$OR$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$OR$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$OR$CI[1],5), ", ", signif(x$estimates$OR$CI[2],5), ")", sep=""),"\n") 
			cat("\n              log(OR): ", signif(x$estimates$OR$log.psi,5))
			cat("\n    variance(log(OR)): ", signif(x$estimates$OR$var.log.psi,5),"\n")

		}
 
  } else {
  	 stop("Error calling print.summary.tmle. 'x' needs to have class 'summary.tmle'\n")
  }
}

#-------------print.tmle------------------
# print object returned by tmle function
#-----------------------------------------
print.tmle <- function(x,...) {
	if(identical(class(x), "tmle")){
		if(!is.null(x$estimates$ATE)){
			cat(" Additive Effect")
			cat("\n   Parameter Estimate: ", signif(x$estimates$ATE$psi,5))
			cat("\n   Estimated Variance: ", signif(x$estimates$ATE$var.psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$ATE$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$ATE$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$ATE$CI[1],5), ", ", signif(x$estimates$ATE$CI[2],5), ")", sep=""),"\n") 
		}
		if(!is.null(x$estimates$RR)){
			cat("\n Relative Risk")
			cat("\n   Parameter Estimate: ", signif(x$estimates$RR$psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$RR$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$RR$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$RR$CI[1],5), ", ", signif(x$estimates$RR$CI[2],5), ")", sep=""),"\n") 
			cat("\n              log(RR): ", signif(x$estimates$RR$log.psi,5))
			cat("\n    variance(log(RR)): ", signif(x$estimates$RR$var.log.psi,5),"\n")

		}
		if(!is.null(x$estimates$OR)){
			cat("\n Odds Ratio")
			cat("\n   Parameter Estimate: ", signif(x$estimates$OR$psi,5))
			cat("\n              p-value: ", ifelse(x$estimates$OR$pvalue <= 2*10^-16, "<2e-16",signif(x$estimates$OR$pvalue,5)))
			cat("\n    95% Conf Interval:",paste("(", signif(x$estimates$OR$CI[1],5), ", ", signif(x$estimates$OR$CI[2],5), ")", sep=""),"\n") 
			cat("\n              log(OR): ", signif(x$estimates$OR$log.psi,5))
			cat("\n    variance(log(OR)): ", signif(x$estimates$OR$var.log.psi,5),"\n")

		}
	} else {
 		stop("Error calling print.tmle. 'x' needs to have class 'tmle'\n")
 }}

#-------------.verify_args------------------
# initial checks on data passed in
#-------------------------------------------	 
.verify_args <- function(Y,A,W,Delta,Q){
	ok <- c(length(Y) == length(A) & length(A) == NROW(W) & length(A)==NROW(Delta),
	 	 	all(A[!is.na(A)] %in% 0:1),
			sum(is.na(A), is.na(W), is.na(Delta)) == 0,
			is.null(Q)
			)
	if(!ok[4]) {ok[4] <- ncol(Q)==2 & nrow(Q)==length(A)}
	warning_messages <- c("\t'Y', 'A', 'W', 'Delta', must contain the same number of observations\n",
				"\t'A' must be binary (0,1)\n",
				"\tNo missing values allowed in 'A' or 'W'\n",
				"\tQ must be an nx2 matrix, with columns Q(0,W), Q(1,W)"
				)
	if(!all(ok)){
		warning("\n", warning_messages[!ok])
	}
	return(all(ok))
}


#---------- function .bound ---------------
# set outliers to min/max allowable values
# assumes x contains only numerical data
#-----------------------------------------
.bound <- function(x, bounds){
	x[x>max(bounds)] <- max(bounds)
	x[x<min(bounds)] <- min(bounds)
	return(x)
}


#----- function .estQ_cvSL ----
# purpose: Obtain cross-validated estimates for initial Q using discrete SL.  
# 	This function will return cross-validated predicted initial values 
#   corresponding to the best algorithm in the library, or the convex combination
#   calculated by SL itself, as chosen by cvRSS.
#   The fitted value for each observation i, is predicted from a fit based on a training set
#	excluding the fold containing observation i. Observations with same id are grouped in the
#   same fold.
# arguments: 
# 	Y - outcome
# 	X - design matrix with A in first column
# 	SL.library - prediction algorithms
# 	V - number of outer cv-folds for disscrete SL
# 	V_SL - number of folds for internal SL cross-validation
# 	family - binomial or gaussian
#   Delta - missingness indicator
#   Qbounds - bounds on predicted values for Q
#   id - subject identifier
# returns:
#  Q - nx3 matrix of predicted values on linear scale (e.g. logit if Qfamily=binomial)
#-----------------------------------------------
.estQ_cvSL <- function(Y,X,SL.library=NULL, V=5, V_SL=5, family="gaussian", Delta, Qbounds, id, verbose){
	Q <- cvRSS <- NULL
	n <- length(Y)
	u.id <- unique(id)
	n.id <- length(u.id)
	fold <- by(sample(1:n.id),rep(1:V, length.out=n.id),function(x){which(id %in% u.id[x])})
	n_predictors <-length(SL.library)
	
	if(NCOL(X) > 1 & require(SuperLearner)){
		newX <- rbind(X,X,X)
		newX[(n+1):(2*n),1] <- 1
		newX[(2*n+1):(3*n),1] <- 0
		# We'll create a matrix of predictions - one column for each predictor in the library
		# plus one more for SL itself, with 3*n predicted values per column, corresponding to newX.
   		predictions <- matrix(nrow=3*n, ncol=length(SL.library)+1)
   		m_SL <- NULL
    	for (v in 1:V) {
    		fold_rows <- c(fold[[v]], fold[[v]]+n, fold[[v]]+2*n)
    		if(class(m_SL) != "try-error"){
    			train.observed <- (1:n)[-fold[[v]]][Delta[-fold[[v]]]==1]
    			suppressWarnings(
    				m_SL <- try(SuperLearner(Y=Y[train.observed], X=X[train.observed,], newX=newX[fold_rows,],
    			 	V=V_SL, save.fit.library=FALSE, family=family,SL.library=SL.library,id=id[train.observed]))
    			 )
    		}
    		if(class(m_SL) != "try-error"){
    			predictions[fold_rows,1] <- m_SL$SL.predict
    			for (s in 1:n_predictors){
    				predictions[fold_rows,s+1] <- m_SL$library.predict[,s]
    			}
    			predictions <- .bound(predictions, Qbounds)
    		}
    	} 
    	cvRSS <- colSums(Delta*(Y-predictions[1:n,])^2)
    	best <- which.min(cvRSS)
    	best_alg <- c("SL", SL.library)[best]
    	Q <- matrix(data=predictions[,best], nrow=n, ncol=3, byrow=FALSE)
    	colnames(Q) <- c("QAW", "Q1W", "Q0W")
    } 	
	if(verbose){cat("\tDiscrete SL: best algorithm = ", best_alg,"\n")}
	if (is.null(Q) | class(m_SL) == "try-error"){
		Q <- 0
		class(Q) <- "try-error"
	}
    return(Q)
 }

#-----------estimate_Q----------------
# purpose: estimate Q=E(Y | A,W) data-adaptively,
# unless super learner not available, or user specifies 
# initial values or a regression formula
# arguments: 
# 	Y - outcome 
#	A - treatment indicator (1=treatment, 0=control)
# 	W - baseline covariates
#   Delta - missingness indicator
#	Q - optional externally estimated values for Q
#   Qform - optional regression formula to use for glm if 
#	        non-data adaptive estimation specified
# 	maptoYstar - if TRUE, using logistic fluctuation for bounded, continuous outcomes
# 		estimation inital Q on linear scale, bounded by (0,1),and return on logit scale
# 	SL.library - library of prediction algorithms for Super Learner
# 	family - binomial or gaussian
#   id - subject identifier
# returns matrix of linear predictors for Q(A,W), Q(1,W), Q(0,W), 
#		family for stage 2 targeting
#		coef, NA, unless Q is estimated using a parametric model
# 		type, estimation method for Q
#----------------------------------------
estimate_Q <- function (Y,A,W, Delta, Q, Qbounds, Qform, maptoYstar, 
		SL.library, cvQinit, family, id, verbose) {

	Qfamily <- family
	m <- NULL
	coef <- NA
	type <- "user-supplied values"
	if(is.null(Q)){
		if(verbose) { cat("\tEstimating initial regression of Y on A and W\n")}
		Q <- matrix(NA, nrow=length(Y), ncol = 3)
  	  	colnames(Q)<- c("QAW", "Q1W", "Q0W")
  	  	
  	  	if(!(is.null(Qform))){
  	  		m <- glm(Qform, data=data.frame(Y,A,W, Delta), family=family, subset=Delta==1)
  	  		Q[,"QAW"] <- predict(m, newdata=data.frame(Y,A,W), type="response")
  	  		Q[,"Q1W"] <- predict(m, newdata=data.frame(Y,A=1,W), type="response")
  	  		Q[,"Q0W"] <- predict(m, newdata=data.frame(Y,A=0,W), type="response")
  	  		coef <- coef(m)
  	  		type="glm, user-supplied model"	
  	  	} else {
  	  		if(cvQinit){
  	  			m <- try(.estQ_cvSL(Y,X=cbind(A,W),SL.library, family=family, Delta=Delta, Qbounds=Qbounds,id=id, verbose=verbose))
  	  			if(class(m) != "try-error"){
  	  				type <- "cross-validated SL"
  	  				Q <- m
  	  			}
  	  		} else {
  	  			if(require(SuperLearner)){
  					if(verbose) {cat(" using SuperLearner\n")}
  					n <- length(Y)
  					X <- data.frame(A,W)
  					X1 <- data.frame(A=1, W)
  					X0 <- data.frame(A=0, W)
  					m <- try(SuperLearner(Y[Delta==1],X[Delta==1,], newX=rbind(X, X1, X0), SL.library=SL.library,
  							V=5, family=family, save.fit.library=FALSE, id=id[Delta==1]))
  					if(class(m) == "SuperLearner"){
  						Q[,"QAW"] <- m$SL.predict[1:n]
  						Q[,"Q1W"] <- m$SL.predict[(n+1):(2*n)]
  						Q[,"Q0W"] <- m$SL.predict[(2*n+1):(3*n)]
  						type <- "SL"
  	  	} } } }
  	} 
  	if(is.na(Q[1,1]) | identical(class(m), "try-error")){
  	  		if(verbose) {cat("\t Running main terms regression for 'Q' using glm\n")}
  	  		m <- glm(Y~., data=data.frame(Y,A,W, Delta), family=family, subset=Delta==1)
  	  		Q[,"QAW"] <- predict(m, newdata=data.frame(Y,A,W), type="response")
  	  		Q[,"Q1W"] <- predict(m, newdata=data.frame(Y,A=1,W), type="response")
  	  		Q[,"Q0W"] <- predict(m, newdata=data.frame(Y,A=0,W), type="response")
  	  		coef <- coef(m)
  	  		type="glm, main terms model"
  	 }
	Q <- .bound(Q, Qbounds)
	if(maptoYstar | identical(Qfamily,"binomial") | identical(Qfamily, binomial)){
			Q <- qlogis(Q)
			Qfamily <- "binomial"	
	}
	return(list(Q=Q, family=Qfamily, coef=coef, type=type))
}

#-----------estimate_g----------------
# Estimate g1W=P(A=1|W) or P(Delta=1|A,W)
# d - dataframe (A,W) or (Delta, A,W) when estimating missingness
# g1W - optional vector of externally estimated values
# gform - optionalformula to use for glm
# SL.library - algorithms to use for super learner estimation
#   id - subject identifier
# verbose - flag, whether or not to print messages
# message - printed when verbose=TRUE 
#----------------------------------------
estimate_g <- function (d, g1W, gform,SL.library, id, verbose, message)  {
  m <- NULL
  coef <- NA
  type <- NULL
  if (is.null(g1W)){
  	if(verbose){cat("\tEstimating", message, "\n")}
	if (all(d[,1]==1)) {  	# true when estimating Delta nad there's no missingness
		g1W <- rep(1,nrow(d))
		type <- "No missingness"
	} else {
  	  if (is.null(gform)){
  		if(require(SuperLearner)){
  			m <- try(SuperLearner(Y=d[,1], X=d[,-1], family="binomial", SL.library=SL.library,
  							V=5, save.fit.library=FALSE, id=id))
  			if(class(m) == "SuperLearner") {
  				g1W <- m$SL.predict
  			}
  	    } else {
  			if(verbose){cat("\tRunning main terms regression for 'g' using glm\n")}
			form <- paste(colnames(d)[1],"~1", paste(colnames(d)[-1], collapse = "+"), sep="+")  
			m <- glm(form, data=d, family="binomial")
			g1W <- predict(m, type="response")
			coef <- coef(m)
  		}
  	  } else {
  	  	form <- try(as.formula(gform))
  	  	if(class(form)== "formula") {
  	  		m <- try(glm(form,  data=d, family="binomial"))
  	  		if (class(m)[1]=="try-error"){
  				if(verbose){cat("\tInvalid formula supplied. Running glm using main terms\n")}
				form <- paste(colnames(d)[1],"~1", paste(colnames(d)[-1], collapse = "+"), sep="+")  
  	  			m <- glm(form, data=d, family="binomial")
  	  		} else {
  	  			type <- "user-supplied regression formula"
  	  		}
  	  	} else {
  	  	  	if(verbose){cat("\tRunning main terms regression for 'g' using glm\n")}
			form <- paste(colnames(d)[1],"~1", paste(colnames(d)[-1], collapse = "+"), sep="+")  
  	  		m <- glm(form, data=d, family="binomial")
  	  	}
  	  	 g1W <- predict(m, newdata=d, type="response")
  	  	 coef <- coef(m)
  	  }
  	}} else {
  		type <- "user-supplied values"
  	}
  	if(is.null(type)){ type <- class(m)[1]}
 return(list(g1W=g1W, coef=coef, type=type))
}


#----------------------- calc_parameters -----------------------
# Calculate parameter estimates, CIs and p-values
# arguments:
#   Y - outcome
#   A - binary treatment indicator
# 	g1W - values of P(A=1|W)*P(Delta=1|A,W)
# 	g0W - values of P(A=0|W)*P(Delta=1|A,W)
#   Delta =- missingness indicator 
#   Q - nx3 matrix(Q(A,W), Q(1,W), Q(0,W))
#   mu1 - targeted estimate of EY1
#   mu0 - targeted estimate of EY0
#   id - subject id
#   family - "gaussian" or "binomial" 
# returns:
#   list: (ATE, RR, OR) 
#	  ATE - additive effect (risk difference for binary outcomes)
#	  RR - relative risk (NULL if family = gaussian)
#	  OR - odds ratio (NULL if family = gaussian)
#     each of these has elements psi, CI, pvalue
#     ATE also has var.psi, RR and OR also have log.psi and var.log.psi
#--------------------------------------------------------------
calc_parameters <- function(Y,A, Delta, g1W, g0W, Q, mu1, mu0, id, family){
	n.id <- length(unique(id))
	ATE <- RR <- OR <- NULL
	Y[is.na(Y)] <- 0
	IC.ATE<- (A/g1W - (1-A)/g0W)*Delta*(Y-Q[,"QAW"]) + Q[,"Q1W"] - Q[,"Q0W"] - (mu1-mu0)
	IC.ATE <- as.vector(by(IC.ATE, id, mean))
	ATE$psi <- mu1-mu0
	ATE$var.psi <- var(IC.ATE)/n.id
	ATE$CI <- ATE$psi + c(-1.96,1.96)*sqrt(ATE$var.psi)
	ATE$pvalue <- 2*pnorm(-abs(ATE$psi/sqrt(ATE$var.psi)))

	if(family=="binomial"){		
		IC.logRR <- 1/mu1*((A/g1W)*Delta*(Y-Q[,"QAW"]) + Q[,"Q1W"] - mu1) - 
					1/mu0*((1-A)/g0W*Delta*(Y-Q[,"QAW"])+ Q[,"Q0W"] - mu0)
		IC.logRR <- as.vector(by(IC.logRR, id, mean))
		var.psi.logRR <- var(IC.logRR)/n.id
		RR$psi <- mu1/mu0
		RR$CI  <- exp(log(RR$psi) + c(-1.96,1.96)*sqrt(var.psi.logRR))
		RR$pvalue <- 2*pnorm(-abs(log(RR$psi)/sqrt(var.psi.logRR)))
		RR$log.psi <- log(RR$psi)
		RR$var.log.psi <- var.psi.logRR

		IC.logOR <- 1/(mu1*(1-mu1)) * (A/g1W*Delta*(Y-Q[,"QAW"]) + Q[,"Q1W"]) - 
					1/(mu0*(1-mu0)) * ((1-A)/g0W*Delta * (Y-Q[,"QAW"]) + Q[,"Q0W"])
		IC.logOR <- as.vector(by(IC.logOR, id, mean))
		var.psi.logOR <- var(IC.logOR)/n.id
		OR$psi <-  mu1/(1-mu1)/(mu0 / (1-mu0))
		OR$CI  <- exp(log(OR$psi) + c(-1.96,1.96)*sqrt(var.psi.logOR))
		OR$pvalue <- 2*pnorm(-abs(log(OR$psi)/sqrt(var.psi.logOR)))
		OR$log.psi <- log(OR$psi)
		OR$var.log.psi <- var.psi.logOR
	}
	return(list(ATE=ATE, RR=RR, OR=OR))
}

#-------------------------------tmle----------------------------------------
# estimate marginal treatment effect for binary point treatment
# accounting for missing outcomes.
# arguments:
# Y - outcome
# A - binary treatment indicator, 1-treatment, 0 - control
# W - vector, matrix or dataframe containing baseline covariates
# Delta - indicator of missingness.  1 - observed, 0 - missing
# Q - E(Y|A,W), optional matrix of values containing two columns 1) E(Y|A=0,W), 2) E(Y|A=1,W) 
# g1W - optional values for P(A=1|W)
# gform - optional glm regression formula 
# gbound - one value for symmetrical bounds on g1W, or a vector containing upper and lower bounds
# pDelta1 - optional values for P(Delta=1|A,W)
# g_Deltaform - optional glm regression formula  
# Q.SL.library- optional library for estimation of Q, 
#		defaults to DSA, glm, step, knn if family = "binomial", DSAM glm, step , ridge, otherwise
# cvQinit - if TRUE obtain cross-validated initial Q with super learner
# g.SL.library - optional library for estimation of g and g_Delta
#		defaults to DSA, glm, step, knn (last three included with base R installation)
# family - family specification for regression models, defaults to gaussian
# fluctuation - "logistic" (default) or "linear" (for targeting step)
# alpha - bound on predicted probabilities (0.005, 0.995 default)
# id - optional subject identifier
# verbose - flag for controlling printing of messages
#	Returns
#	psi, treatment effect estimate,
#   pvalue, 2-sided p-value 
#	var - estimated variance of parameter estimate,
#	epsilon - coefficient used in targeting step
#	coefficients and predicted values for Q_n^0(A,W), g(1,W), g_Delta(1,A,W)
#-------------------------------------------------------------------------------
tmle <- function(Y,A,W,Delta=rep(1,length(Y)),  
				Q=NULL, Qform=NULL, Qbounds=NULL, 
				g1W=NULL, gform=NULL, gbound=0.025, 
				pDelta1=NULL, g_Deltaform=NULL, 
				Q.SL.library=NULL, cvQinit=FALSE,
				g.SL.library=c("SL.glm", "SL.step", "SL.knn", "SL.DSA.2"),
				family="gaussian", fluctuation="logistic",
				alpha  = 0.995, id=1:length(Y),verbose=FALSE) {
	psi.tmle <- varIC <- CI <- pvalue <- NA
	if(!.verify_args(Y,A,W,Delta,Q)){
		stop()
	}
	
	if (identical(family,binomial)) {family == "binomial"}
	if (identical(family,gaussian)) {family == "gaussian"}
	if(is.null(Q.SL.library)){
   		if(family=="binomial"){
   			Q.SL.library <- c("SL.glm", "SL.step", "SL.knn", "SL.DSA.2")
   		} else {
   			Q.SL.library <- c("SL.glm", "SL.step", "SL.DSA.2")
   		}
   	}
   	if(is.null(colnames(W))){ colnames(W) <- paste("W",1:ncol(W), sep="")}
	invalid.name <- nchar(colnames(W)) == 0
	if(any(invalid.name)){
		colnames(W)[invalid.name] <- paste(".internalW", which(invalid.name), sep="")
	}

 	maptoYstar <- fluctuation=="logistic"
 		
 	if(family=="binomial") {Qbounds <- c(0,1)}
 	if(is.null(Qbounds)) { 
 		Qbounds <- range(Y)
   	}
   	if(!is.null(Q)){
   		QAW <- (1-A)*Q[,1] + A*Q[,2]
   		Q <- cbind(QAW, Q1W=Q[,2], Q0W=Q[,1])
   	} 		
   	ab <- c(0,1)
   	Ystar <- Y
   	if(maptoYstar){ 
   		Ystar <- .bound(Y, Qbounds)
   		if(!is.null(Q)){
   			Q <- .bound(Q, Qbounds)
   		}
   		if(0 >= alpha | 1 <= alpha){
			alpha <- .995
			warning(paste("\n\talpha must be between 0 and 1, alpha reset to",alpha,"\n"))
		}
		ab <- range(Ystar, na.rm=TRUE)
		Ystar[is.na(Ystar)] <- 0  
		Ystar <- (Ystar-ab[1])/diff(ab)	
		if(!is.null(Q)){Q <- (Q-ab[1])/diff(ab)}
		Qbounds <- c(alpha, 1-alpha)	
	} 
 				
	Q <- estimate_Q(Y=Ystar,A,W, Delta, Q, Qbounds, Qform, 
					maptoYstar=maptoYstar, SL.library=Q.SL.library, 
					cvQinit=cvQinit, family=family, id=id, verbose=verbose)
	if(length(gbound)==1){gbound <- c(gbound, 1-gbound)}
 	g <- estimate_g(d=data.frame(A,W), g1W, gform, g.SL.library, id=id, verbose, "treatment mechanism") 
 	g_Delta <- estimate_g(d=data.frame(Delta, A, W), pDelta1, g_Deltaform, 
 	 				g.SL.library,id=id, verbose, "missingness mechanism") 
 	
  	g1W <- g$g1W 
  	g1W.total <- .bound(g$g1W*g_Delta$g1W, gbound)
  	g0W.total <- .bound((1-g$g1W)*g_Delta$g1W, gbound)
  	h1W <- A/g1W.total
  	h0W <- (1-A)/g0W.total
  	suppressWarnings(
  		epsilon <- coef(glm(Ystar~-1 + offset(Q$Q[,"QAW"]) + h1W + h0W, family=Q$family, subset=Delta==1))
  	)
 	Qstar <- Q$Q + c((epsilon[1]*h1W + epsilon[2]*h0W), epsilon[1]/g1W.total, epsilon[2]/g0W.total)
	colnames(Qstar) <- c("QAW", "Q1W", "Q0W")
 
   	if (maptoYstar | identical(Q$family,"binomial")) {
		Qstar <- plogis(Qstar)*diff(ab)+ab[1] 
		Q$Q <- plogis(Q$Q)*diff(ab)+ab[1]
		Ystar <- Ystar*diff(ab)+ab[1]
	}
   	res <- calc_parameters(Ystar, A, Delta, g1W.total, g0W.total, Qstar, 
   	   		  mu1=mean(Qstar[,"Q1W"]), mu0=mean(Qstar[,"Q0W"]), id, family)
  	returnVal <- list(estimates=res, Qinit=Q, g=g, g_Delta=g_Delta, Qstar=Qstar) 
  	class(returnVal) <- "tmle"
  	return(returnVal)
}
